{-----------------------------------------------------------------------------
 Unit Name: Fparam4
 Project Name : Parametricas4
 Author:    Angel Montesinos Amilibia
 Dates: 2003-2009
 Purpose: Implements class 'TParam4' which is the main form of the application,
    with the menu
-----------------------------------------------------------------------------}
{

Parametricas4, workshop for the geometry of surfaces in R^4.


MIT License

Copyright (c) [2003 - 2018] [Angel Montesinos-Amilibia], email montesin@uv.es

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.


}


unit Fparam4;

interface

uses
  SysUtils, Windows, Graphics, Messages, Classes, Controls, Forms, Dialogs,
  StdCtrls, Buttons, ExtCtrls, StdActns, ActnList, Menus, Printers, ComCtrls,
  VectorsAndUtil, Grids, ValEdit;

const

  {tags in menus}

  kParabolicLocus = 0;
  kSemiumbilicalLocus = 1;
  kHParallelEllipseLocus = 2;
  kZeroGKLocus = 3;
  kZeroDiscrNormDevLocus = 4;
  kZeroDiscrLatDevLocus = 5;
  kZeroDiscrStrongPpalLocus = 6;
  kEraseLoci = 14;

  {tags for background color}
  kColorBy0InEllipse = 0;
  kColorByBxC = 1;
  kColorByAngleHE = 2;
  kColorByGK = 3;
  kDiscrNormDev = 4;
  kDiscrLatDev = 5;
  kDiscrStrPpal = 6;
  kEraseBackground = 17;

  {tags for types of lines}
  kCoordLines = 0;
  kCurvatureLines = 1;
  kAsymptoticLines = 2;
  kMeanCurvLines = 3;
  kPrincipalLines = 4;
  kExtrNormDevLines = 5;
  kExtrLatDevLines = 6;
  kStrongPrincipalLines = 7;
  kEraseLines = 15;

  kShowFoliation1 = 16;
  kShowFoliation2 = 17;

  {End tags in menus}

  kFilter = 'pr4'; {goes in InputStrs[-3]}
  kCreator = 'Angel Montesinos-Amilibia'; {goes in InputStrs[-2]}
  kVersion000 = '0'; {goes in InputStrs[-1]}
  kLastStr = 8; {goes in InputStrs[0]. InputStrs go from -3 to kLastStr. These
  strings are the data to save.}

{Geometry of the panels used in the form}

  kWidthMainBox = 548;
  kHeightMainBox = 548;
  kWidthEllipseBox = 224;
  kHeightEllipseBox = 240;

  {Miscellaneous}
  kReturn = char(13);
  kEnter = char(3);
  kEscape = char(27);
  kTab = char($09);

type

  wireKnot = record
    key: byte;
    pos: VectorsAndUtil.Vector2;
  end;

  wireKnots = array of wireKnot;

  wireData = record
    text: AnsiString;
    last: Integer;
  end;

  Curve = record
    c: Tcolor;
    last: Integer;
    pts: array of VectorsAndUtil.Vector2;
  end;

  CurvesArray = record
    last: Integer;
    cc: array of Curve;
  end;

  PtArray = record
    last: Integer;
    pts: array of VectorsAndUtil.Vector2;
  end;

  Pltt = array[0..255] of TColor;
  PalettePtr = ^Pltt;

  TParam4 = class(TForm)
    xEdit: TEdit;
    Label1: TLabel;
    Label4: TLabel;
    yEdit: TEdit;
    Label5: TLabel;
    zEdit: TEdit;
    Label6: TLabel;
    wEdit: TEdit;
    Label7: TLabel;
    uMinEdit: TEdit;
    uMaxEdit: TEdit;
    Label8: TLabel;
    vMinEdit: TEdit;
    vMaxEdit: TEdit;
    Label9: TLabel;
    EnterButton: TSpeedButton;
    MainMenu1: TMainMenu;
    File1: TMenuItem;
    Open1: TMenuItem;
    Save1: TMenuItem;
    SaveAs1: TMenuItem;
    Print1: TMenuItem;
    N1: TMenuItem;
    Exit1: TMenuItem;
    Edit1: TMenuItem;
    Cut1: TMenuItem;
    Copy1: TMenuItem;
    Paste1: TMenuItem;
    Loci1: TMenuItem;
    ParabolicLocus1: TMenuItem;
    SemiumbilicalLocus1: TMenuItem;
    HEllipseLocus1: TMenuItem;
    N2: TMenuItem;
    Foliations1: TMenuItem;
    N3: TMenuItem;
    CurvatureLines1: TMenuItem;
    AsymptoticLines1: TMenuItem;
    MeanCurvatureLines1: TMenuItem;
    EditCut1: TEditCut;
    EditCopy1: TEditCopy;
    EditPaste1: TEditPaste;
    CoordinateLines1: TMenuItem;
    ActionList1: TActionList;
    OpenDialog1: TOpenDialog;
    SaveDialog1: TSaveDialog;
    PrintDialog1: TPrintDialog;
    VanishingGK: TMenuItem;
    EraseLoci1: TMenuItem;
    N6: TMenuItem;
    Erasetheselines1: TMenuItem;
    Ppal1: TMenuItem;
    MainBox: TPaintBox;
    EllipseBox: TPaintBox;
    InfoBar: TStatusBar;
    Manual1: TMenuItem;
    N7: TMenuItem;
    Backgrounds1: TMenuItem;
    ColorbyOrigininEllipse1: TMenuItem;
    ColorbyBxC1: TMenuItem;
    ColorbyangleHMajorAxis1: TMenuItem;
    ColorbyGaussCurv1: TMenuItem;
    N5: TMenuItem;
    EraseBackground1: TMenuItem;
    View1: TMenuItem;
    ColorPalette1: TMenuItem;
    FunctionValues1: TValueListEditor;
    N4: TMenuItem;
    ShowFoliationOne1: TMenuItem;
    ShowFoliationTwo1: TMenuItem;
    ExtremalNormalDeviationLines1: TMenuItem;
    ExtremalLateralDeviationLines1: TMenuItem;
    ZeroDiscrNormalDeviation1: TMenuItem;
    ZeroDiscrLateralDeviation1: TMenuItem;
    Discrmaxnormaldeviation1: TMenuItem;
    Discrmaxlateraldeviation1: TMenuItem;
    StrongPrincipalLines1: TMenuItem;
    ZeroDiscrStrongPpalLines1: TMenuItem;
    ColorbyDiscrStrongPpalDirs1: TMenuItem;

    procedure AddAndDrawLineSegment(vec1, vec2: VectorsAndUtil.Vector2; key: Byte);
    procedure AddAndDrawLineSegmentExtremals(vec1, vec2: VectorsAndUtil.Vector2; key: Byte; cl:
      Integer);
    procedure AddLocusSegment(currentX: VectorsAndUtil.Vector2; key: Integer);
    procedure BeginWatch;
    procedure ColorByFunction(bm: TBitmap);
    procedure ComputeEllipse;
    procedure DoNoEvent(X, Y: Integer; can: TCanvas);
    procedure DrawCoordinateLines(can: TCanvas; param: VectorsAndUtil.Vector2);
    procedure DrawCross;
    procedure DrawEllipse;
    procedure DrawFocalLocus;
    procedure DrawLines(can: TCanvas; r: TRect);
    procedure DrawLocus(can: TCanvas; r: TRect; toPrint: Boolean; locusType:
      Integer);
    procedure DrawOrthogonalVector(firstPt, lastPt: TPoint; param: VectorsAndUtil.Vector2);
    procedure DrawSegment(vec1, vec2: VectorsAndUtil.Vector2; clr: TColor);
    procedure EndWatch;
    function EraseLastLine: Boolean;
    function FromDisk: AnsiString;
    function FromR2toWindow(param: VectorsAndUtil.Vector2): TPoint;
    function FromWindowToR2(x, y: Integer): VectorsAndUtil.Vector2;
    procedure HandleThreadDone(sender: TObject);
    procedure InitializeLines;
    procedure InitializeLoci;
    function InRange(var depth: Double; p: VectorsAndUtil.Vector2): Boolean;
    procedure ManageMenus;
    procedure Manual1Click(Sender: TObject);
    procedure MoveInEllipsePanel(Sender: TObject; Shift: TShiftState; X,
      Y: Integer);
    procedure OnCloseQuery(Sender: TObject; var CanClose: Boolean);
    procedure OnColorByFunction(Sender: TObject);
    procedure OnCreateForm(Sender: TObject);
    procedure OnDestroyForm(Sender: TObject);
    procedure OnEnterButton(Sender: TObject);
    procedure OnEraseBackground(Sender: TObject);
    procedure OnEraseCurves(Sender: TObject);
    procedure OnEraseLoci(Sender: TObject);
    procedure OnExit(Sender: TObject);
    procedure OnKeyDownInForm(Sender: TObject; var Key: Word;
      Shift: TShiftState);
    procedure OnLineItem(Sender: TObject);
    procedure OnLocusItem(Sender: TObject);
    procedure OnMouseDown(Sender: TObject; Button: TMouseButton;
      Shift: TShiftState; X, Y: Integer);
    procedure OnMouseMove(Sender: TObject; Shift: TShiftState; X,
      Y: Integer);
    procedure OnMouseMoveInForm(Sender: TObject; Shift: TShiftState; X,
      Y: Integer);
    procedure OnMouseUp(Sender: TObject; Button: TMouseButton;
      Shift: TShiftState; X, Y: Integer);
    procedure OnOpen(Sender: TObject);
    procedure OnPaint(Sender: TObject);
    procedure OnPrint(Sender: TObject);
    procedure OnSave(Sender: TObject);
    procedure OnSaveAs(Sender: TObject);
    procedure OnShowFoliations(Sender: TObject);
    procedure OnViewPalette(Sender: TObject);
    procedure PaintEllipsePanel(Sender: TObject);
    function ReadData: Boolean;
    procedure RedrawLines;
    procedure RedrawLoci;
    procedure ShowData(param: VectorsAndUtil.Vector2; halfMajorAxis, halfMinorAxis, lengthH,
      depth, alpha,
      KGAuss, discrNormDev, discrLatDev, discrStrPpal: Double);
    procedure ToDisk;

  private
    firstPt, lastPt: TPoint;
    cX, cY, qx, qy: Double;
    crossParam: VectorsAndUtil.Vector2;
    drawOrthonLine, lockPtData, dirty: Boolean;
    decSeparator:Char;

  public
    { Public declarations }
    bm: TBitmap;
    InputStrs: array[-3..kLastStr] of AnsiString;
    palette: PalettePtr;
    rctMain, rctEllipse: TRect;
    min2D, max2D, edges2D, halfEdges2D, center2D: VectorsAndUtil.Vector2;
    e1, e2: Vector4;
    sc, b1, c1, b2, c2, h, h0, v0, e1h, e1b, e1c, e2h, e2b, e2c,
      centerX, centerY, originX, originY, ellFactor: Double;
    margin, curvesParts, numOfCPU, thrNumber: Integer;
    span2D, radius2D, maxSegLen: Double;
    busy, alreadyParsed, stop, crossDrawn, PrintEquations,
      PrintBackground, printingBackground, PrintSubText: Boolean;
    prTitle, prSubText: AnsiString;
    prResolution: Integer;
    cross: array[1..5] of ScreenSegment;
    Filename: TFilename;
    loci: array[kParabolicLocus..kZeroDiscrStrongPpalLocus] of WireKnots;
    lociData: array[kParabolicLocus..kZeroDiscrStrongPpalLocus] of WireData;
    linesOption, lociOption: Integer; {According with the tags in the const section
                                       "tags for types of lines"}

    lines: CurvesArray;
    coordLines: PtArray;
    showFoliations: Byte;
    backgroundType: Integer; {-1 means no background, other values refer to tags
    in the const section "tags for background colors"}

  end;

var
  Param4: TParam4;

implementation

{$R *.DFM}

uses

  Math,
  FEvaluatorBytes_64_XE2, FunctionParserBytes_64_XE2, ResourceStrings, FParametrizationR2R4, DiffEqSolver,
  LinesToPrintDlg, FImplicitCurveLine, FManParam4, FYesNotCancel, FLinePalette,
  BackgroundThreads, Printing, FunctionAndPartDerivs_64;

function ClipLine(x, v: VectorsAndUtil.Vector2; r: TRect): VectorsAndUtil.Vector2;
{given the line   t -> x + t v  and the rect 'r',
then the line is inside 'r' when 't' is greater
than Result[1] and less than Result[2]. If Result[1]>
Result[2], the line does not cut 'r'}
var
  t1, t2: VectorsAndUtil.Vector2;
  a: Double;
begin
  Result[1]:= 1;
  Result[2]:= 0;
  if abs(v[1]) > 1.0e-8 then begin
    t1[1]:= (r.left - x[1]) / v[1];
    t1[2]:= (r.right - x[1]) / v[1];
    if t1[1] > t1[2] then begin {swap them}
      a:= t1[1];
      t1[1]:= t1[2];
      t1[2]:= a;
    end;
    if abs(v[2]) > 1.0e-8 then begin
      t2[1]:= (r.top - x[2]) / v[2];
      t2[2]:= (r.bottom - x[2]) / v[2];
      if t2[1] > t2[2] then begin {swap them}
        a:= t2[1];
        t2[1]:= t2[2];
        t2[2]:= a;
      end;
      Result[1]:= Max(t1[1], t2[1]);
      Result[2]:= Min(t1[2], t2[2]);
    end else if (x[2] >= r.top) and (x[2] < r.bottom) then
      Result:= t1;
  end else if (x[1] >= r.left) and (x[1] < r.right) and (abs(v[2]) > 1.0e-8) then
    begin
    Result[1]:= (r.top - x[2]) / v[2];
    Result[2]:= (r.bottom - x[2]) / v[2];
    if Result[1] > Result[2] then begin {swap them}
      a:= Result[1];
      Result[1]:= Result[2];
      Result[2]:= a;
      Exit;
    end;
  end;
end;

procedure TParam4.AddAndDrawLineSegment(vec1, vec2: VectorsAndUtil.Vector2; key: Byte);
{if 'key=1' initiate a new arc. Add 'vec2' to the current arc, and draw
 the segment from 'vec1' to 'vec2'}
var
  pt: TPoint;
  i: Integer;
begin
  if key = 1 then begin
    inc(lines.last);
    if high(lines.cc) < lines.last then begin
      SetLength(lines.cc, lines.last + 5);
      for i:= 1 to 4 do
        lines.cc[lines.last + i].last:= -1;
    end;
    lines.cc[lines.last].last:= -1;
    lines.cc[lines.last].c:= LinePalette.curveColor;
  end;
  inc(lines.cc[lines.last].last);
  if high(lines.cc[lines.last].pts) < lines.cc[lines.last].last then
    SetLength(lines.cc[lines.last].pts, lines.cc[lines.last].last + 100);
  lines.cc[lines.last].pts[lines.cc[lines.last].last]:= vec2;
  MainBox.Canvas.Pen.Color:= lines.cc[lines.last].c;
  pt:= FromR2ToWindow(vec1);
  MainBox.canvas.moveto(pt.x, pt.y);
  pt:= FromR2ToWindow(vec2);
  MainBox.canvas.lineto(pt.x, pt.y);
end;

procedure TParam4.AddAndDrawLineSegmentExtremals(vec1, vec2: VectorsAndUtil.Vector2; key: Byte;
  cl: Integer);
{similar to AddAndDrawLineSegment but now the color of the arc is chosen
 according with the class of line, i.e whether it represents a maximal direction,
 or minimal or relative maximal or relative minimal}
var
  pt: TPoint;
  i: Integer;
begin
  if key = 1 then begin
    inc(lines.last);
    if high(lines.cc) < lines.last then begin
      SetLength(lines.cc, lines.last + 5);
      for i:= 1 to 4 do
        lines.cc[lines.last + i].last:= -1;
    end;
    lines.cc[lines.last].last:= -1;
    case cl of
      1: lines.cc[lines.last].c:= {clFuchsia} $00DD00DD;
      2: lines.cc[lines.last].c:= {clAqua} $00FF8800;
      3: lines.cc[lines.last].c:= {clLime} $0000DD00;
      4: lines.cc[lines.last].c:= clRed;
    else
      lines.cc[lines.last].c:= LinePalette.curveColor;
    end;
  end;
  inc(lines.cc[lines.last].last);
  if high(lines.cc[lines.last].pts) < lines.cc[lines.last].last then
    SetLength(lines.cc[lines.last].pts, lines.cc[lines.last].last + 100);
  lines.cc[lines.last].pts[lines.cc[lines.last].last]:= vec2;
  MainBox.Canvas.Pen.Color:= lines.cc[lines.last].c;
  pt:= FromR2ToWindow(vec1);
  MainBox.canvas.moveto(pt.x, pt.y);
  pt:= FromR2ToWindow(vec2);
  MainBox.canvas.lineto(pt.x, pt.y);
end;

procedure TParam4.AddLocusSegment(currentX: VectorsAndUtil.Vector2; key: Integer);
{add 'currentX' to 'lociData' with its key}
var
  i: Integer;
begin
  inc(lociData[lociOption].last);
  i:= lociData[lociOption].last;
  if i > High(loci[lociOption]) then
    SetLength(loci[lociOption], i + 50);
  loci[lociOption, i].pos:= currentX;
  loci[lociOption, i].key:= key;
end;

procedure TParam4.BeginWatch;
{the caption of 'EnterButton' is put to 'eStop'. User can
 then click it and stop the calculation of a background or of
 some curves}
begin
  busy:= True;
  stop:= False;
  EnterButton.Font.Color:= clRed;
  EnterButton.Caption:= eStop;
  EnterButton.Update;
  cursor:= crHourGlass;
end;

procedure Tparam4.ColorByFunction(bm: TBitmap);
{fill 'bm' with a color that represents the value of some function}
var
  cu, cv, qh, qv: Double;
  w, h, hMin, hMax, i: Integer;
begin
  beginWatch;
  if (thrNumber = 0) then begin {all threads must have ended before beginning
    the same task}
    SetLength(thr, numOfCPU);
      {give room for as many threads as the number of CPU}

    {set coefficients for passing between bitmap and domain coordinates}
    w:= bm.width;
    h:= bm.height;
    cU:= (w - 1) / edges2D[1];
    cV:= (h - 1) / edges2D[2];
    qh:= -cu * min2D[1];
    qv:= cv * max2D[2];

    {erase 'bm' contents}
    bm.canvas.FillRect(Rect(0, 0, w, h));

    {create first thread}
    hMin:= 0;
    hMax:= h div numOfCPU;
    thr[0]:= TBackgroundThread.CreateBT(True, cu, cv, qh, qv, bm,
      hMin, hMax - 1, 0);

    {create more threads if necessary for doing the drawing of the rows
     from 'hMin' to 'hMax-1'}
    if numOfCPU > 0 then
      for i:= 1 to numOfCPU - 1 do begin
        hMin:= hMax;
        hMax:= h * (i + 1) div numOfCPU;
        thr[i]:= TBackgroundThread.CreateBT(True, cu, cv, qh, qv, bm,
          hMin, hMax - 1, i);
      end;
    try
      for i:= 0 to numOfCPU - 1 do begin {initialize some fields of the threads}
        thr[i].FreeOnTerminate:= True;
        thr[i].Priority:= tpHigher;
        thr[i].OnTerminate:= HandleThreadDone;
        thr[i].tag:= i;
      end;
      thrNumber:= 0;
      for i:= 0 to numOfCPU - 1 do begin {launch the threads}
        thr[i].Resume;
        inc(thrNumber);
      end;
    except
      for i:= 0 to numOfCPU - 1 do
        thr[i].Free;
      thrNumber:= 0;
      endWatch;
    end;
  end else begin
    for i:= 0 to numOfCPU - 1 do
      if Assigned(thr[i]) then begin
        thr[i].Suspend;
        thr[i].Free;
      end;
  end;
end;

procedure TParam4.ComputeEllipse;
{We define an orthonormal basis of the normal space such that its first
vector is parallel to the mean curvature vector. Then, compute the components
of B and C in that basis}
var
  aux: Double;
  m: matrix44;
  i: Integer;
begin
  h:= MapR2R4.nh;  {norm of mean curvature vector}
  if h > 1.0e-7 then begin
    aux:= 1 / h;
    for i:= 1 to 4 do
      e1[i]:= MapR2R4.vH[i] * aux;
  end else if MapR2R4.nb > 1.0e-7 then begin {if H =0, use B instead}
    aux:= 1 / MapR2R4.nb;
    for i:= 1 to 4 do
      e1[i]:= MapR2R4.vB[i] * aux;
  end else if MapR2R4.nc > 1.0e-7 then begin  {if B = 0, use C instead}
    aux:= 1 / MapR2R4.nc;
    for i:= 1 to 4 do
      e1[i]:= MapR2R4.vC[i] * aux;
  end else begin {there is nothing to draw}
    sc:= 0;
    Exit;
  end;
  m[1]:= MapR2R4.t;
  m[2]:= MapR2R4.Jt;
  m[3]:= e1;
  e2:= crossProd(m); {this gives the second basis vector}

  {components of B and C}
  b1:= EuclInnerProd(e1, MapR2R4.vB);
  b2:= EuclInnerProd(e2, MapR2R4.vB);
  c1:= EuclInnerProd(e1, MapR2R4.vC);
  c2:= EuclInnerProd(e2, MapR2R4.vC);
end;

procedure TParam4.DoNoEvent(X, Y: Integer; can: TCanvas);
{mouse is in pixel [x,y]: show data about that point and if necessary
 draw the directions defined at that point}
var
  param: VectorsAndUtil.Vector2;
  halfMajorAxis, halfMinorAxis, longH, depth, alpha, KGAuss, discrNormDev,
    discrLatDev, discrStrPpal: Double;
  i: Integer;
begin
  if alreadyParsed then
    param:= FromWindowToR2(x, y) {get point of domain represented by pixel [x,y]}
  else begin {clear fields of function values box}
    for i:= 0 to FunctionValues1.Strings.Count - 1 do
      FunctionValues1.Strings.Strings[i]:= ' = ';
    InfoBar.SimpleText:= '';
    cursor:= crDefault;
    Exit;
  end;
  if (linesOption = kCoordLines) and MapR2R4.ComputeInvariants(param,
    halfMajorAxis,
    halfMinorAxis, longH, depth, alpha, KGAuss, discrNormDev, discrLatDev, discrStrPpal) then
      begin
    {show data about point, write ellipsoid axes
         and draw ellipsoid and ellipse in the corresponding window}

    ShowData(param, halfMajorAxis, halfMinorAxis, longH, depth, alpha, KGAuss,
      discrNormDev, discrLatDev, discrStrPpal);
    ComputeEllipse;
    DrawEllipse;
  end else begin
    if crossDrawn then
      DrawCross; {erase previous cross}
    if MapR2R4.ComputeCross(param, linesOption, halfMajorAxis,
      halfMinorAxis, longH, depth, alpha, KGAuss, discrNormDev, discrLatDev, discrStrPpal) then
        begin
      DrawCross;
      crossParam:= param;
      ShowData(param, halfMajorAxis, halfMinorAxis, longH, depth, alpha, KGAuss,
        discrNormDev, discrLatDev, discrStrPpal);
      ComputeEllipse;
      DrawEllipse;
    end else {perhaps there are no valid directions at the point. Show data}
       if MapR2R4.ComputeInvariants(param, halfMajorAxis, halfMinorAxis, longH,
            depth, alpha, KGAuss, discrNormDev, discrLatDev, discrStrPpal) then
        begin
      ShowData(param, halfMajorAxis, halfMinorAxis, longH, depth, alpha, KGAuss,
        discrNormDev, discrLatDev, discrStrPpal);
      ComputeEllipse;
      DrawEllipse;
    end;
  end
end;

procedure TParam4.DrawCoordinateLines(can: TCanvas; param: VectorsAndUtil.Vector2);
{draw the coordinate lines that pass by 'param'}
var
  pt: TPoint;
begin
  MainBox.canvas.Pen.Color:= $00CCCCCC;
  inc(CoordLines.last);
  if High(CoordLines.pts) < CoordLines.last then
    SetLength(CoordLines.pts, CoordLines.last + 4);
  CoordLines.pts[CoordLines.last]:= param;
  pt:= FromR2ToWindow(param);
  MainBox.canvas.moveto(0, pt.y);
  MainBox.canvas.lineto(rctMain.Right, pt.y);
  MainBox.canvas.moveto(pt.x, 0);
  MainBox.canvas.lineto(pt.x, rctMain.bottom);
end;

procedure TParam4.DrawCross;
{draw the segments that indicate the directions, at the last point
 where they were computed, of the class chosen by user in the menu}
var
  num, i: Integer;
begin
  MainBox.canvas.Pen.Color:= linePalette.CurveColor;
  num:= curvesParts div 2;
  if num > 0 then begin
    MainBox.canvas.Pen.Mode:= pmNotXor;
    for i:= 1 to num do begin
      MainBox.canvas.moveto(cross[i, 1].x, cross[i, 1].y);
      MainBox.canvas.lineto(cross[i, 2].x, cross[i, 2].y);
    end;
    crossDrawn:= not crossDrawn;
  end;
end;

procedure TParam4.DrawEllipse;
{Draw the "Null Gauss Curvature circle, curvature ellipse, the focal locus, the mean curvature vector, etc.
 The ellipse is drawn so that the mean curvature vector appears as horizontal
 and the center of ellipse at right of the origin of the normal plane}
const
  cnum = 24;
var
  str: AnsiString;
  rd, t, si, co, step, minA, maxA: Double;
  i, x, y: Integer;
  can: TCanvas;
begin
  can:= ellipseBox.Canvas;
  can.FillRect(rctEllipse);

  {center the drawing in the window and scale the drawing}

  rd:= sqrt(MapR2R4.bb + MapR2R4.cc);  {radius of Null Gauss curvature
     circle. It is a circle centered at the center of ellipse. The
     Gauss curvature of the surface at the point vanishes iff the origin,
     that is the other end of the mean curvature vector is on that circle}

  maxA:= h + rd; {maximum horizontal abscissa of the drawing}
  minA:= min(0, h - rd); {minimum horizontal abscissa of the drawing}
  t:= max(maxA - minA, 2 * rd); {horizontal amplitude must comprise at least the circle}
  if t > 1.0e-9 then begin {else, don't draw anything}
    sc:= 0.9 * rctEllipse.right / t; {give a margin of ten percent}
    centerY:= rctEllipse.bottom div 2;  {center of ellipse must be at half height of panel}
    originY:= rctEllipse.bottom / 2;  {the same for the origin}
    if h > rd then begin
      originX:= 0.05 * rctEllipse.right; {origin is the leftmost point}
      centerX:= originX + sc * h; {panel abscissa of ellipse center}
    end else begin
      centerX:= rctEllipse.right / 2; {center of ellipse at center of panel}
      originX:= centerX - sc * h;  {panel abscissa for the origin}
    end;
    v0:= centerY;
    h0:= originX;

    {draw mean curvature vector}
    can.Pen.Color:= $0000FF00;
    can.MoveTo(round(centerX), round(centerY));
    can.LineTo(round(originX), round(centerY));

    {draw ellipse}
    t:= 0;
    step:= 2 * pi / cNum;
    sincos(t, si, co);
    x:= round(centerX + sc * (b1 * co + c1 * si));
    y:= round(centerY + sc * (b2 * co + c2 * si));
    can.Pen.Color:= clBlue;
    can.MoveTo(x, y);
    for i:= 1 to cnum do begin
      t:= t + step;
      sincos(t, si, co);
      x:= round(centerX + sc * (b1 * co + c1 * si));
      y:= round(centerY + sc * (b2 * co + c2 * si));
      can.LineTo(x, y);
    end;

    DrawFocalLocus;

    {draw null Gauss curvature circle}
    x:= round(centerX + sc * rd);
    y:= round(centerY);
    can.Pen.Color:= clAqua;
    can.moveto(x, y);
    for i:= 1 to cnum do begin
      t:= t + step;
      sincos(t, si, co);
      x:= round(centerX + sc * rd * co);
      y:= round(centerY + sc * rd * si);
      can.LineTo(x, y);
    end;

    {draw a horizontal segment and its length as a measure of the scale of
     the figure}
    can.Pen.Color:= clSilver;
    margin:= round(0.05 * rctEllipse.right);
    can.moveto(margin, margin div 2);
    str:= FormatFloatString(0.45 * rctEllipse.right / sc, 7);
    can.lineto(rctEllipse.right div 2, margin div 2);
    can.TextOut(rctEllipse.right div 2 + 5, 2, str);
  end
end;

procedure Tparam4.DrawFocalLocus;
{draw the intersection of the Focal Locus, as a subset of the normal bundle,
 with the fibre of that bundle over the point. If is a conic, drawn in red.
 If the point is semiumbilical it degenerates to a pair of straight lines.
 See the note "FocalSetOfSurfaces.pdf" in this same folder}
const
  cNum = 100;
var
  vComps, JvComps, s: VectorsAndUtil.Vector2;
  t, u, v, up, vp, etapetap, etaetap, lambda, step, si, co, scale: Double;
  pt: TPoint;
  can: TCanvas;
  i: Integer;
  goodPast, goodCurrent, signPast, signCur: Boolean;
  r: TRect;
begin
  scale:= sc * ellFactor; {sometimes the focal locus is too far to be visible in
   the panel. User may diminish its scale to make it visible by changing 'ellipFactor'}
  can:= ellipseBox.Canvas;
  can.Pen.Color:= clRed;
  if abs(b1 * c2 - c1 * b2) < 1.0e-6 then begin {semiumbilic case. Then the focal
  locus consists of two segments.}
    if max(mapR2R4.nb, mapR2R4.nc) < 1.0e-8 then
      exit;
    lambda:= sqrt(mapR2R4.bb + mapR2R4.cc);

    {vectors v1 and Jv1}
    if mapR2R4.nb > mapR2R4.nc then begin
      vComps[1]:= h +  lambda*b1/mapR2R4.nb;
      vComps[2]:= lambda*b2/mapR2R4.nb;
    end
    else begin
      vComps[1]:= h + lambda*c1/mapR2R4.nc;
      vComps[2]:= lambda*c2/mapR2R4.nc;
    end;
    JvComps[1]:= -vComps[2];
    JvComps[2]:= vComps[1];

    u:= sqr(vComps[1])+ sqr(vComps[2]);
    if u < 1.0e-16 then
      Exit;
    {inverse of v1 projected to window}
    vComps[1]:= originX + scale*vComps[1]/u;
    vComps[2]:= originY + scale*vComps[2]/u;

    {straight line passing by vComps}
    s:= ClipLine(vComps, JvComps, rctEllipse);
    if s[1] < s[2] then begin
      pt.X:= round(vComps[1] + s[1] * JvComps[1]);
      pt.Y:= round(vComps[2] + s[1] * JvComps[2]);
      can.MoveTo(pt.x, pt.y);
      pt.X:= round(vComps[1] + s[2] * JvComps[1]);
      pt.Y:= round(vComps[2] + s[2] * JvComps[2]);
      can.LineTo(pt.x, pt.y);
    end;

   {vectors v2 and Jv2}
   if mapR2R4.nb > mapR2R4.nc then begin
      vComps[1]:= h -  lambda*b1/mapR2R4.nb;
      vComps[2]:= - lambda*b2/mapR2R4.nb;
    end
    else begin
      vComps[1]:= h - lambda*c1/mapR2R4.nc;
      vComps[2]:= - lambda*c2/mapR2R4.nc;
    end;
    JvComps[1]:= -vComps[2];
    JvComps[2]:= vComps[1];

    u:= sqr(vComps[1])+ sqr(vComps[2]);
    if u < 1.0e-16 then
      Exit;
    {inverse of v2 projected to window}
    vComps[1]:= originX + scale*vComps[1]/u;
    vComps[2]:= originY + scale*vComps[2]/u;

    {straight line passing by vComps}
    s:= ClipLine(vComps, JvComps, rctEllipse);
    if s[1] < s[2] then begin
      pt.X:= round(vComps[1] + s[1] * JvComps[1]);
      pt.Y:= round(vComps[2] + s[1] * JvComps[2]);
      can.MoveTo(pt.x, pt.y);
      pt.X:= round(vComps[1] + s[2] * JvComps[1]);
      pt.Y:= round(vComps[2] + s[2] * JvComps[2]);
      can.LineTo(pt.x, pt.y);
    end;
  end else begin {usual case}
    i:= rctEllipse.right - rctEllipse.left;
    r.Left:= -100 * i;
    r.Right:= 101 * i;
    i:= rctEllipse.bottom - rctEllipse.top;
    r.Top:= -3 * i;
    r.Bottom:= 4 * i;
    step:= 2 * pi / cNum;
    t:= 0;
    sincos(t, si, co);

    {first point in the ellipse}
    u:= h + b1 * co + c1 * si;
    v:= b2 * co + c2 * si;

    {derivative of eta at that point}
    up:= -b1 * si + c1 * co;
    vp:= -b2 * si + c2 * co;

    {computation of the inverted pedal point}
    etapetap:= sqr(up) + sqr(vp);
    etaetap:= u * up + v * vp;
    lambda:= -etaetap / etapetap;
    u:= u + lambda * up;
    v:= v + lambda * vp;
    signCur:= u * vp - v * up > 0;
    etapetap:= sqr(u) + sqr(v);
    goodCurrent:= (etapetap > 1.0e-8);
    if goodCurrent then begin
      u:= u / etapetap;
      v:= v / etapetap;
      pt.x:= round(originX + scale * u);
      pt.y:= round(originy + scale * v);
      goodCurrent:= ptInRect(r, pt);
      if goodCurrent then
        can.moveto(pt.x, pt.y);
    end;

    {the same with the remaining points of the ellipse}
    for i:= 1 to cNum do begin
      t:= i * step;
      goodPast:= goodCurrent;
      signPast:= signCur;
      sincos(t, si, co);
      u:= h + b1 * co + c1 * si;
      v:= b2 * co + c2 * si;
      up:= -b1 * si + c1 * co;
      vp:= -b2 * si + c2 * co;
      etapetap:= sqr(up) + sqr(vp);
      etaetap:= u * up + v * vp;
      lambda:= -etaetap / etapetap;
      u:= u + lambda * up;
      v:= v + lambda * vp;
      etapetap:= sqr(u) + sqr(v);
      goodCurrent:= etapetap > 1.0e-8;
      signCur:= u * vp - v * up > 0;
      if goodCurrent then begin
        u:= u / etapetap;
        v:= v / etapetap;
        pt.x:= round(originX + scale * u);
        pt.y:= round(originy + scale * v);
        goodCurrent:= ptInRect(r, pt);
        if goodPast and goodCurrent and (signCur = signPast) then
          can.lineto(pt.x, pt.y)
        else if goodCurrent then
          can.moveto(pt.x, pt.y);
      end;
    end;
  end;
end;

procedure TParam4.DrawLines(can: TCanvas; r: TRect);
{draw to window the computed lines}
var
  scX, scY, offX, offY: Double;
  i, j: Integer;
  pixel: TPoint;
  prevPen: TPen;
begin
  if not alreadyParsed then
    Exit;
  scX:= (r.right - r.left) / edges2D[1];
  scY:= (r.bottom - r.top) / edges2D[2];
  offX:= (r.right + r.left) / 2 - scX * center2D[1];
  offy:= (r.bottom + r.top) / 2 + scY * center2D[2];
  prevPen:= can.Pen;
  can.Pen.Mode:= pmCopy;
  can.Pen.Color:= $00CCCCCC;
  if CoordLines.last >= 0 then for i:= 0 to CoordLines.last do begin
      pixel.x:= round(scX * CoordLines.pts[i, 1] + offX);
      pixel.y:= round(-scY * CoordLines.pts[i, 2] + offY);
      can.moveto(r.left, pixel.y);
      can.lineto(r.Right, pixel.y);
      can.moveto(pixel.x, r.top);
      can.lineto(pixel.x, r.bottom);
    end;
  if (lines.last >= 0) and (lines.last <= High(lines.cc)) then
    for i:= 0 to lines.last do begin
      can.Pen.Color:= lines.cc[i].c;
      if (lines.cc[i].last > 0) and (lines.cc[i].last <= High(lines.cc[i].pts))
        then begin
        pixel.x:= round(scX * lines.cc[i].pts[0, 1] + offX);
        pixel.y:= round(-scY * lines.cc[i].pts[0, 2] + offY);
        can.moveto(pixel.x, pixel.y);
        for j:= 1 to lines.cc[i].last do begin
          pixel.x:= round(scX * lines.cc[i].pts[j, 1] + offX);
          pixel.y:= round(-scY * lines.cc[i].pts[j, 2] + offY);
          can.lineto(pixel.x, pixel.y);
        end;
      end;
    end;
  can.Pen:= prevPen;
end;

procedure TParam4.DrawLocus(can: TCanvas; r: TRect; toPrint: Boolean; locusType:
  Integer);
{Draw the pertinent locus to window}
var
  scX, scY, offX, offY: Double;
  i: Integer;
  pixel: TPoint;
begin
  scX:= (r.right - r.left) / edges2D[1];
  scY:= (r.bottom - r.top) / edges2D[2];
  offX:= (r.right + r.left) / 2 - scX * center2D[1];
  offy:= (r.bottom + r.top) / 2 + scY * center2D[2];
  can.pen.color:= clLtGray;
  can.Pen.Mode:= pmCopy;
  if not toPrint then
    can.Pen.Width:= 1;
  if lociData[locusType].last > 1 then begin
    i:= 0;
    pixel.x:= round(scX * loci[locusType, i].pos[1] + offX);
    pixel.y:= round(-scY * loci[locusType, i].pos[2] + offY);
    can.moveto(pixel.x, pixel.y);
    for i:= 1 to lociData[locusType].last do begin
      pixel.x:= round(scX * loci[locusType, i].pos[1] + offX);
      pixel.y:= round(-scY * loci[locusType, i].pos[2] + offY);
      if loci[locusType, i].key = 1 then
        can.moveto(pixel.x, pixel.y)
      else
        can.lineto(pixel.x, pixel.y);
    end;
  end;
end;

procedure TParam4.DrawOrthogonalVector(firstPt, lastPt: TPoint;
{Draw vector orthogonal and of same length as that signalled by
 the mouse by means of the pixels 'firstPt, lastPt'}
  param: VectorsAndUtil.Vector2);
var
  det: Double;
  sh, orth: VectorsAndUtil.Vector2;
  tip: TPoint;
begin
  MapR2R4.ComputeFff(param);
  sh[1]:= (lastPt.x - firstPt.x) / cX;
  sh[2]:= (firstPt.y - lastPt.y) / cY;
  det:= sqrt(abs(MapR2R4.guu * MapR2R4.gvv - sqr(MapR2R4.guv)));
  if det < 1.0E-10 then
    Exit;
  orth[1]:= -(MapR2R4.guv * sh[1] + MapR2R4.gvv * sh[2]) / det;
  orth[2]:= (MapR2R4.guu * sh[1] + MapR2R4.guv * sh[2]) / det;
  tip.x:= firstPt.x + round(orth[1] * cX);
  tip.y:= firstPt.y - round(orth[2] * cY);
  MainBox.canvas.moveto(firstPt.x, firstPt.y);
  MainBox.canvas.lineto(tip.x, tip.y);
end;

procedure TParam4.DrawSegment(vec1, vec2: VectorsAndUtil.Vector2; clr: TColor);
var
  punto: TPoint;
begin
  MainBox.canvas.Pen.Color:= clr;
  punto:= FromR2ToWindow(vec1);
  MainBox.canvas.moveto(punto.x, punto.y);
  punto:= FromR2ToWindow(vec2);
  MainBox.canvas.lineto(punto.x, punto.y);
end;

procedure TParam4.EndWatch;
{a lengthy computation has ended. Change button accordingly}
begin
  busy:= False;
  stop:= True;
  EnterButton.Font.Color:= clNavy;
  EnterButton.Caption:= eEnter;
  EnterButton.Update;
  cursor:= crDefault;
end;

function TParam4.EraseLastLine: Boolean;
{when user press <ctrl> + Z, mark last line stored part as non existent}
begin
  result:= False;
  if (lines.last >= 0) and (lines.last <= high(lines.cc)) then begin
    lines.cc[lines.last].last:= -1;
    lines.cc[lines.last].pts:= nil;
    dec(lines.last);
    result:= True;
  end;
end;

function TParam4.FromDisk: AnsiString;
{read data for a surface from disk}
var
  sl: TStringList;
begin
  result:= '';
  sl:= TStringList.Create;
  try
    sl.LoadFromFile(fileName);
    try
      if (sl.Strings[0] <> kFilter) or (sl.Strings[1] <> kCreator) then
        raise Exception.Create(eWrongFile);
      if (sl.Strings[2] = kVersion000) then begin
        if StrToInt(sl.Strings[3]) <> kLastStr then
          raise Exception.Create(eCorruptedFile);
        xEdit.Text:= sl.Strings[4];
        yEdit.Text:= sl.Strings[5];
        zEdit.Text:= sl.Strings[6];
        wEdit.Text:= sl.Strings[7];
        uMinEdit.Text:= sl.Strings[8];
        uMaxEdit.Text:= sl.Strings[9];
        vMinEdit.Text:= sl.Strings[10];
        vMaxEdit.Text:= sl.Strings[11];
      end else
        raise Exception.Create(eVersionUnsupported);
    except on E: Exception do begin
        xEdit.Text:= sl.Strings[0];
        yEdit.Text:= sl.Strings[1];
        zEdit.Text:= sl.Strings[2];
        wEdit.Text:= sl.Strings[3];
        uMinEdit.Text:= sl.Strings[4];
        uMaxEdit.Text:= sl.Strings[5];
        vMinEdit.Text:= sl.Strings[6];
        vMaxEdit.Text:= sl.Strings[7];
      end;
    end;
  finally
    FreeAndNil(sl);
  end;
end;

function TParam4.FromR2toWindow(param: VectorsAndUtil.Vector2): TPoint;
begin
  result.x:= round(cX * param[1] + qX);
  result.y:= round(-cY * param[2] + qY);
end;

function TParam4.FromWindowToR2(x, y: Integer): VectorsAndUtil.Vector2;
begin
  result[1]:= (x - qX) / cX;
  result[2]:= (qy - y) / cY;
end;

procedure TParam4.HandleThreadDone(sender: TObject);
begin
  dec(thrNumber);
  if thrNumber = 0 then begin
    if printingBackground then begin
      PrintAll;
      printBackground:= False;
      printingBackground:= False;
    end else begin
      MainBox.Canvas.Draw(0, 0, bm);
      RedrawLoci;
      RedrawLines;
      endWatch;
    end;
  end;
end;


procedure TParam4.InitializeLines;
var
  i: Integer;
begin
  coordLines.last:= -1;
  lines.last:= -1;
  if high(lines.cc) >= 0 then begin
    for i:= 0 to high(lines.cc) do begin
      lines.cc[i].pts:= nil;
      lines.cc[i].last:= -1;
    end;
    lines.cc:= nil;
  end;
end;

procedure TParam4.InitializeLoci;
var
  i: Integer;
begin
  for i:= kParabolicLocus to kZeroDiscrStrongPpalLocus do begin
    SetLength(loci[i], 100);
    lociData[i].last:= -1;
  end;
end;


function TParam4.InRange(var depth: Double; p: VectorsAndUtil.Vector2): Boolean;
{returns whether the point 'p' is in domain, and in 'depth' gives a
 measure of the depth of the point in or out the domain. depth < 0
 ifthe point is out}
begin
  depth:= min(halfEdges2D[1] - abs(p[1] - center2D[1]),
    halfEdges2D[2] - abs(p[2] - center2D[2]));
  result:= depth >= 0;
end;

procedure TParam4.ManageMenus;
begin
  if Busy then
    Exit;
  Save1.Enabled:= alreadyParsed and dirty;
  SaveAs1.Enabled:= alreadyParsed;
  Print1.Enabled:= alreadyParsed;
  Loci1.Enabled:= alreadyParsed;
  Backgrounds1.Enabled:= alreadyParsed;
  Foliations1.Enabled:= alreadyParsed;
  ColorbyOrigininEllipse1.Checked:= (backgroundType =
    ColorbyOrigininEllipse1.tag);
  ColorbyBxC1.Checked:= (backgroundType = ColorbyBxC1.tag);
  ColorbyangleHMajorAxis1.Checked:= (backgroundType =
    ColorbyangleHMajorAxis1.tag);
  ColorbyGaussCurv1.Checked:= (backgroundType = ColorbyGaussCurv1.tag);
  Discrmaxnormaldeviation1.Checked:= (backgroundType = kDiscrNormDev);
  Discrmaxlateraldeviation1.checked:= (backgroundType = kDiscrLatDev);
  ColorbyDiscrStrongPpalDirs1.Checked:= (backgroundType = kDiscrStrPpal);
  CoordinateLines1.checked:= (linesOption = kCoordLines);
  CurvatureLines1.checked:= (linesOption = kCurvatureLines);
  AsymptoticLines1.checked:= (linesOption = kAsymptoticLines);
  MeanCurvatureLines1.checked:= (linesOption = kMeanCurvLines);
  ExtremalNormalDeviationLines1.Checked:= (linesOption = kExtrNormDevLines);
  ExtremalLateralDeviationLines1.Checked:= (linesOption = kExtrLatDevLines);
  StrongPrincipalLines1.Checked:= (linesOption = kStrongPrincipalLines );
  Ppal1.Checked:= (linesOption = kPrincipalLines);
  ShowFoliationOne1.Enabled:= (linesOption <> kStrongPrincipalLines);
  ShowFoliationTwo1.Enabled:= ShowFoliationOne1.Enabled;
end;

procedure TParam4.Manual1Click(Sender: TObject);
{open manual window}
begin
  if busy then
    Exit;
  if not Assigned(Manual) then
    Manual:= TManual.Create(nil);
  Manual.Show;
end;

procedure TParam4.MoveInEllipsePanel(Sender: TObject;
  Shift: TShiftState; X, Y: Integer);
{when lockPtData, show the curvatures of the u-second
 fundamental form of the surface when 'u' is determined
 by the mouse position in the panel}
var
  nu: Vector4;
  str: AnsiString;
  u, v, mi, ma, aux, len2, n11, n12, n22, KK, HH, bn, cn: Double;
  p1, p2, pp1, pp2: VectorsAndUtil.Vector2;
  pt: TPoint;
  i: Integer;
begin
  if not (lockPtData and alreadyParsed) or busy or (sc = 0) then
    Exit;
  DrawEllipse;

  {obtain coordinates in the normal space of point signalled by the mouse}
  u:= (x - h0) / sc;
  v:= (y - v0) / sc;

  len2:= 0;

  {compute the normal vector defined by mouse. Remember that e1 and e2 is
   an orthonormal basis of the normal space}
  for i:= 1 to 4 do begin
    nu[i]:= u * e1[i] + v * e2[i];
    len2:= len2 + sqr(nu[i]);
  end;

  if len2 > 1.0e-7 then begin
    len2:= 1 / sqrt(len2);
    for i:= 1 to 4 do {normalize 'nu'}
      nu[i]:= nu[i] * len2;

    {project orthogonally the second fundamental form on the line determined by 'nu'
     compute its Gauss and mean curvatures, and write them in the panel}
    n11:= EuclInnerProd(nu, MapR2R4.G11);
    n12:= EuclInnerProd(nu, MapR2R4.G12);
    n22:= EuclInnerProd(nu, MapR2R4.G22);
    HH:= (n11 + n22) / 2;
    KK:= n11 * n22 - sqr(n12);
    EllipseBox.canvas.Pen.Color:= clSilver;
    str:= 'K= ' + FormatFloatString(KK, 8) + ';  H= ' + FormatFloatString(HH,
      8);
    EllipseBox.canvas.TextOut(round(0.05 * kWidthEllipseBox), kWidthEllipseBox,
      str);

    {draw the segments that illustrate the preceding operation}
    aux:= hypot(u, v);
    u:= u / aux;
    v:= v / aux;
    ma:= aux;
    mi:= 0;

    bn:= u * b1 + v * b2;
    cn:= u * c1 + v * c2;
    aux:= 1 / hypot(bn, cn);
    bn:= bn * aux;
    cn:= cn * aux;
    p1[1]:= h + bn * b1 + cn * c1;
    p1[2]:= bn * b2 + cn * c2;
    p2[1]:= h - bn * b1 - cn * c1;
    p2[2]:= -bn * b2 - cn * c2;

    n11:= -v * p1[1] + u * p1[2];
    pp1[1]:= p1[1] + n11 * v;
    pp1[2]:= p1[2] - n11 * u;
    n11:= p1[1] * u + p1[2] * v;
    ma:= max(ma, n11);
    mi:= min(mi, n11);

    n12:= -v * p2[1] + u * p2[2];
    pp2[1]:= p2[1] + n12 * v;
    pp2[2]:= p2[2] - n12 * u;
    n12:= p2[1] * u + p2[2] * v;
    ma:= max(ma, n12);
    mi:= min(mi, n12);

    with EllipseBox.canvas do begin
      pt.x:= round(h0 + sc * p1[1]);
      pt.y:= round(v0 + sc * p1[2]);
      MoveTo(pt.x, pt.y);
      pt.x:= round(h0 + sc * pp1[1]);
      pt.y:= round(v0 + sc * pp1[2]);
      LineTo(pt.x, pt.y);

      pt.x:= round(h0 + sc * p2[1]);
      pt.y:= round(v0 + sc * p2[2]);
      MoveTo(pt.x, pt.y);
      pt.x:= round(h0 + sc * pp2[1]);
      pt.y:= round(v0 + sc * pp2[2]);
      LineTo(pt.x, pt.y);

      pt.x:= round(h0 + sc * mi * u);
      pt.y:= round(v0 + sc * mi * v);
      Moveto(pt.x, pt.y);
      pt.x:= round(h0 + sc * ma * u);
      pt.y:= round(v0 + sc * ma * v);
      LineTo(pt.x, pt.y)
    end;
  end;
end;

procedure TParam4.OnCloseQuery(Sender: TObject; var CanClose: Boolean);
var
  res: Word;
begin
  if busy then
    Exit;
  if dirty then begin
    with TYesNotCancel.Create(nil) do  {give user chance to save data}
      try
        res:= ShowModal;
      finally
        Release;
      end;
    if res = mrNo then
      CanClose:= True
    else if res = mrYes then begin
      OnSave(sender);
      CanClose:= not dirty;
    end else
      CanClose:= False;
  end else
    CanClose:= True;
end;

procedure TParam4.OnColorByFunction(Sender: TObject);
{fill background with a color gamma according with the value of some
function}
begin
  if Busy or (not alreadyParsed) then
    Exit;
  printingBackground:= False;
  backgroundType:= TMenuItem(sender).tag;
  ManageMenus;
  ColorByFunction(bm);
end;

procedure TParam4.OnCreateForm(Sender: TObject);
var
  i, j: Integer;
begin
  try
    rctMain:= Rect(0, 0, kWidthMainBox, kHeightMainBox);
    rctEllipse:= Rect(0, 0, kWidthEllipseBox, kHeightEllipseBox);
    mainBox.canvas.FillRect(rctMain);
    firstPt.x:= 200;
    firstPt.y:= 200;
    lastPt:= firstPt;
    drawOrthonLine:= False;
    lockPtData:= False;
    backgroundType:= -1;
    stop:= False;
    crossDrawn:= False;
    PrintBackground:= False;
    printingBackground:= False;
    PrintEquations:= True;
    PrintSubText:= True;
    alreadyParsed:= False;
    linesOption:= kCoordLines;
    lociOption:= kZeroGKLocus;
    EllipseBox.Canvas.Brush.Color:= clWhite;
    EllipseBox.Canvas.Brush.Style:= bsSolid;
    sc:= 0;
    ellFactor:= 1;
    numOfCPU:= CPUCount;
    bm:= TBitmap.Create;
    bm.PixelFormat:= pf32bit;
    bm.Width:= kWidthMainBox;
    bm.Height:= kHeightMainBox;
    inputStrs[-3]:= kFilter;
    inputStrs[-2]:= kCreator;
    inputStrs[-1]:= kVersion000;
    inputStrs[0]:= IntToStr(kLastStr);
    lines.last:= -1;
    coordLines.last:= -1;
    InitializeLoci;
    New(palette);

    FunctionValues1.Selection:= TGridRect(Rect(20, 20, 20, 20));

    theParser:= TFunctionParserBytes_64.Create;
    SetLength(theFsClones, numOfCPu);
    for i:= 1 to 4 do begin
      theFs[i]:= TEvaluatorBytes_64.Create;
      for j:= 0 to numOfCPu-1 do
        theFsClones[j,i]:= TEvaluatorBytes_64.Create;
    end;

    SetLength(theFPDClones, numOfCPu);
    for i:= 1 to 4 do begin
      theFPD[i]:= TFunctionAndPartDerivs_64.Create;
      for j:= 0 to numOfCPu-1 do
        theFPDClones[j,i]:= TFunctionAndPartDerivs_64.Create;
    end;

    showFoliations:= 3;
    decSeparator:= FormatSettings.decimalSeparator;
    FormatSettings.decimalSeparator:= '.';
  except on Exception do
      Halt;
  end;
end;

procedure TParam4.OnDestroyForm(Sender: TObject);
var
  i, j: Integer;
begin
  FormatSettings.decimalSeparator:= decSeparator;
  bm:= nil;
  for i:= kParabolicLocus to kZeroDiscrStrongPpalLocus do begin
    loci[i]:= nil;
  end;
  lines.last:= -1;
  if High(lines.cc) >= 0 then
    for i:= 0 to High(lines.cc) do begin
      lines.cc[i].last:= 0;
      if high(lines.cc[i].pts) >= 0 then
        lines.cc[i].pts:= nil;
    end;
  if High(coordLines.pts) >= 0 then
    coordLines.pts:= nil;
  theParser.Destroy;
  for i:= 1 to 4 do begin
    theFs[i].Free;
    theFPD[i].Free;
    for j:= 0 to numOfCPu-1 do begin
      theFsClones[j,i].Free;
      theFPDClones[j,i].Free;
    end;
  end;

end;

procedure TParam4.OnEnterButton(Sender: TObject);
var
  i: Integer;
begin
  if Busy then begin
    stop:= True;
    if thrNumber > 0 then
      for i:= 0 to numOfCPU - 1 do
        thr[i].Free
  end else
    ReadData;
end;

procedure TParam4.OnEraseBackground(Sender: TObject);
begin
  if Busy then
    Exit;
  bm.canvas.FillRect(rctMain);
  MainBox.canvas.Draw(0, 0, bm);
  crossDrawn:= False;
  backgroundType:= -1;
  ManageMenus;
  DrawLines(mainBox.canvas, rctMain);
  RedrawLoci;
end;

procedure TParam4.OnEraseCurves(Sender: TObject);
begin
  if Busy or not alreadyParsed then
    Exit;
  MainBox.canvas.FillRect(rctMain);
  crossDrawn:= False;
  InitializeLines;
  MainBox.canvas.Draw(0, 0, bm);
  RedrawLoci;
end;

procedure TParam4.OnEraseLoci(Sender: TObject);
begin
  if Busy or not alreadyParsed then
    Exit;
  MainBox.canvas.FillRect(rctMain);
  crossDrawn:= False;
  MainBox.canvas.Draw(0, 0, bm);
  InitializeLoci;
  DrawLines(mainBox.canvas, rctMain);
end;

procedure TParam4.OnExit(Sender: TObject);
begin
  if Busy then
    Exit;
  OnSave(sender);
  Application.Terminate;
end;

procedure TParam4.OnKeyDownInForm(Sender: TObject; var Key: Word;
  Shift: TShiftState);
begin
  if Busy or not alreadyParsed then
    Exit;
  if ssCtrl in Shift then begin
    if key = ord('Z') then begin
      if EraseLastLine then
        MainBox.Invalidate;
    end else begin
      lockPtData:= not lockPtData;
      if lockPtData then begin {compute components of H,B and C in the orthonormal
         basis  (e1,e2)}
        e1h:= EuclInnerProd(e1, MapR2R4.vH);
        e1b:= EuclInnerProd(e1, MapR2R4.vB);
        e1c:= EuclInnerProd(e1, MapR2R4.vC);
        e2h:= EuclInnerProd(e2, MapR2R4.vH);
        e2b:= EuclInnerProd(e2, MapR2R4.vB);
        e2c:= EuclInnerProd(e2, MapR2R4.vC);
      end;
    end;
  end else if Key in [VK_UP, VK_DOWN] then begin
    {modify the scale for drawing the focal locus}
    if Key = VK_UP then
      ellFactor:= ellFactor * 1.5
    else if Key = VK_DOWN then
      ellFactor:= ellFactor * 0.7;
    if alreadyParsed then
      DrawEllipse;
  end;
end;

procedure TParam4.OnLineItem(Sender: TObject);
begin
  if Busy then
    Exit;
  linesOption:= TMenuItem(sender).Tag;
  if linesOption = kCoordLines then
    curvesParts:= 0
  else if linesOption in [kCurvatureLines, kExtrNormDevLines, kExtrLatDevLines]
    then  {set the number of arcs that start from a point}
    curvesParts:= 8
  else if linesOption = kStrongPrincipalLines then
    curvesParts:= 10
  else
    curvesParts:= 4;
  ManageMenus;
end;

procedure TParam4.OnLocusItem(Sender: TObject);
{draw the locus where some function vanishes}
var
  ff: MapR2R1;
begin
  if Busy or (not alreadyParsed) then
    Exit;
  cursor:= crHourGlass;
  lociOption:= TMenuItem(sender).Tag;
  ManageMenus;
  SetLength(loci[lociOption], 100);
  lociData[lociOption].last:= -1;
  case lociOption of
    kParabolicLocus:
      ff:= MapR2R4.DepthOfOrigin;
    kSemiumbilicalLocus:
      ff:= MapR2R4.JBdotC;
    kHParallelEllipseLocus:
      ff:= MapR2R4.AngleHEllipse;
    kZeroGKLocus:
      ff:= MapR2R4.GaussCurvature;
    kZeroDiscrNormDevLocus:
      ff:= MapR2R4.DiscriminantExtremalNormDevDirs;
    kZeroDiscrLatDevLocus:
      ff:= MapR2R4.DiscriminantExtremalLatDevDirs;
    kZeroDiscrStrongPpalLocus:
      ff:= MapR2R4.DiscriminantStrongPrincipalLines;
  else begin
      Exit;
    end;
  end;
  BeginWatch;
  theImplicitCurveLine.ComputeImplicitCurve(ff, AddLocusSegment,
    min2D, max2D, 40, cMaxNumFlags, radius2D);
  crossDrawn:= False;
  cursor:= crDefault;
  DrawLocus(MainBox.canvas, rctMain, False, lociOption);
  EndWatch;
end;

procedure TParam4.OnMouseDown(Sender: TObject; Button: TMouseButton;
  Shift: TShiftState; X, Y: Integer);
var
  par: VectorsAndUtil.Vector2;
begin
  if Busy or not alreadyParsed then
    Exit;
  InfoBar.SimpleText:= '';
  firstPt.X:= X;
  firstPt.Y:= Y;
  lastPt:= firstPt;
  drawOrthonLine:= ssShift in Shift;
  lockPtData:= False;

  if drawOrthonLine then with MainBox.Canvas do begin
    {begin tracking the mouse with a sticking line}
    Pen.Color:= clLime;
    Pen.Mode:= pmNotXor;
    moveto(X, Y);
    lineto(X, Y);
    Pen.Mode:= pmCopy;
  end;

  {draw lines passing by that point}
  par:= FromWindowToR2(x, y);
  BeginWatch;
  if not drawOrthonLine then begin
    if (linesOption = kCoordLines) then
      DrawCoordinateLines(MainBox.canvas, par)
    else begin
      if crossDrawn then
        DrawCross;
      DrawCurvesByPoint(MainBox.canvas, curvesParts, linesOption, false, par,
        showFoliations);
    end;
  end;
  EndWatch;
end;

procedure TParam4.OnMouseMove(Sender: TObject; Shift: TShiftState; X,
  Y: Integer);
var
  displacement: Integer;
begin
  if busy or lockPtData or not alreadyParsed then
    Exit;
  displacement:= abs(firstPt.x - X) + abs(firstPt.y - Y);
  if (ssLeft in Shift) and (displacement > 0) and alreadyParsed then begin
    if drawOrthonLine then {draw sticking line}
      with mainBox.canvas do begin
        Pen.Mode:= pmXor;
        moveto(firstPt.x, firstPt.y);
        lineto(lastPt.x, lastPt.y);
        moveto(firstPt.x, firstPt.y);
        lineto(x, y);
        Pen.Mode:= pmCopy;
      end;
    lastPt.x:= X;
    lastPt.y:= Y;
  end else if alreadyParsed and not (ssLeft in Shift) and (displacement > 0) then
    doNoEvent(X, Y, mainBox.canvas);
end;

procedure TParam4.OnMouseMoveInForm(Sender: TObject; Shift: TShiftState; X,
  Y: Integer);
begin
  if crossDrawn then
    DrawCross;
end;

procedure TParam4.OnMouseUp(Sender: TObject; Button: TMouseButton;
  Shift: TShiftState; X, Y: Integer);
var
  theOtherPt: TPoint;
  param: VectorsAndUtil.Vector2;
begin
  if lockPtData or Busy or not alreadyParsed then
    Exit;
  EnterButton.caption:= eEnter;
  theOtherPt.x:= X;
  theOtherPt.y:= Y;
  if drawOrthonLine then begin {draw line orthogonal to sticking line}
    mainBox.canvas.Pen.Mode:= pmXor;
    mainBox.canvas.moveto(firstPt.x, firstPt.y);
    mainBox.canvas.lineto(lastPt.x, lastPt.y);
    mainBox.canvas.Pen.Mode:= pmCopy;
    mainBox.canvas.moveto(firstPt.x, firstPt.y);
    mainBox.canvas.lineto(lastPt.x, lastPt.y);
    param:= FromWindowToR2(firstPt.x, firstPt.y);
    DrawOrthogonalVector(firstPt, theOtherPt, param);
    mainBox.canvas.Pen.Mode:= pmCopy;
    drawOrthonLine:= False;
  end;
end;

procedure TParam4.OnOpen(Sender: TObject);
var
  Done: Boolean;
begin
  if Busy then Exit;
  done:= False;
  OpenDialog1.Options:= [ofFileMustExist, ofHideReadOnly, ofNoChangeDir];
  while not Done do begin
    if OpenDialog1.Execute then begin
      if not (ofExtensionDifferent in OpenDialog1.Options) then begin
        fileName:= OpenDialog1.FileName;
        FromDisk;
        ReadData;
        bm.canvas.FillRect(rctMain);
        mainBox.canvas.FillRect(rctMain);
        caption:= ExtractFilename(filename);
        caption:= copy(caption, 1, length(caption) - length(kFilter) - 1);
        dirty:= False;
        ManageMenus;
        Done:= True;
      end
    end else { User cancelled }
      Done:= True;
  end;
end;


procedure TParam4.OnPaint(Sender: TObject);
begin
  MainBox.canvas.Draw(0, 0, bm);
  crossDrawn:= False;
  RedrawLoci;
  DrawLines(mainBox.canvas, rctMain);
end;

procedure TParam4.OnPrint(Sender: TObject);
var
  rslt: integer;
  answer: Byte;
begin
  if Busy or (not alreadyParsed) then
    Exit;
  answer:= mrOK;
  rslt:= Form2.ShowModal;
  if (rslt = mrOK) then begin
    if PrintDialog1.Execute then
      update
    else
      answer:= mrCancel;
    if answer = mrOk then begin
      Busy:= True;
      EnterButton.Font.Color:= clRed;
      EnterButton.Caption:= ePrinting;
      EnterButton.update;
      Printer.orientation:= poPortrait;
      Printer.BeginDoc;
      answer:= PreparePrint;
      if answer <> mrOK then
        InfoBar.SimpleText:= eErrorInPrinting
    end;
  end else
    InfoBar.SimpleText:= ePrintingCancelled;
  EndWatch;
end;

procedure TParam4.OnSave(Sender: TObject);
begin
  if (not alreadyParsed) or Busy or (not dirty) then
    Exit;
  if (filename = '') then
    OnSaveAs(sender)
  else begin
    Busy:= True;
    ToDisk;
    dirty:= False;
    Busy:= False;
  end;

end;

procedure TParam4.OnSaveAs(Sender: TObject);
var
  Done: Boolean;
begin
  if (not alreadyParsed) or Busy then Exit;
  Done:= False;
  while not Done do begin
    if SaveDialog1.Execute then begin
      Busy:= True;
      filename:= SaveDialog1.FileName;
      ToDisk;
      dirty:= False;
      Done:= True;
      Busy:= False;
    end else { User cancelled }
      Done:= True;
  end;
end;

procedure TParam4.OnShowFoliations(Sender: TObject);
var
  tg: Integer;
begin
  tg:= (Sender as TMenuItem).Tag;
  if tg = kShowFoliation1 then
    ShowFoliationOne1.Checked:= not ShowFoliationOne1.Checked
  else if tg = kShowFoliation2 then
    ShowFoliationTwo1.Checked:= not ShowFoliationTwo1.Checked;
  ShowFoliations:= ord(ShowFoliationOne1.Checked) +
    2 * ord(ShowFoliationTwo1.Checked);
end;

procedure TParam4.OnViewPalette(Sender: TObject);
begin
  linePalette.Show;
end;

procedure TParam4.PaintEllipsePanel(Sender: TObject);
begin
  if lockPtData then
    DrawEllipse
  else
    ellipseBox.Canvas.FillRect(rctEllipse);
end;

function TParam4.ReadData: Boolean;
{If result= False, something was wrong, else all was well and everything must
 be initialized}
var
  varTexto, m: ansiString;
  i: Integer;
begin
  result:= False;
  if Busy then
    Exit;
  InfoBar.SimpleText:= '';
  m:= eErrorReadingEditField;
  try
    varTexto:= Trim(xEdit.Text);
    xEdit.Text:= varTexto;
    if (varTexto = '') then
      raise Exception.Create(eBlankField);
    theParser.ParseNewText(varTexto, 'uv', '', False, True, min2D, max2D);
    theFs[1].Build(theParser);
    for i:= 0 to numOfCPu-1 do
      theFsClones[i,1].Build(theParser);
    theFPD[1].ParseNewText(varTexto, 'uv', '', True, True);
    for i:= 0 to numOfCPu-1 do
      theFPDClones[i,1].ParseNewText(varTexto, 'uv', '', True, True);

    inputStrs[1]:= varTexto;

    varTexto:= Trim(yEdit.Text);
    yEdit.Text:= varTexto;
    if (varTexto = '') then
      raise Exception.Create(eBlankField);
    theParser.ParseNewText(varTexto, 'uv', '', False, True, min2D, max2D);
    theFs[2].Build(theParser);
    for i:= 0 to numOfCPu-1 do
      theFsClones[i,2].Build(theParser);
    theFPD[2].ParseNewText(varTexto, 'uv', '', True, True);
    for i:= 0 to numOfCPu-1 do
      theFPDClones[i,2].ParseNewText(varTexto, 'uv', '', True, True);
    inputStrs[2]:= varTexto;

    varTexto:= Trim(zEdit.Text);
    zEdit.Text:= varTexto;
    if (varTexto = '') then
      raise Exception.Create(eBlankField);
    theParser.ParseNewText(varTexto, 'uv', '', False, True, min2D, max2D);
    theFs[3].Build(theParser);
    for i:= 0 to numOfCPu-1 do
      theFsClones[i,3].Build(theParser);
    theFPD[3].ParseNewText(varTexto, 'uv', '', True, True);
    for i:= 0 to numOfCPu-1 do
      theFPDClones[i,3].ParseNewText(varTexto, 'uv', '', True, True);
    inputStrs[3]:= varTexto;

    varTexto:= Trim(wEdit.Text);
    wEdit.Text:= varTexto;
    if (varTexto = '') then
      raise Exception.Create(eBlankField);
    theParser.ParseNewText(varTexto, 'uv', '', False, True, min2D, max2D);
    theFs[4].Build(theParser);
    for i:= 0 to numOfCPu-1 do
      theFsClones[i,4].Build(theParser);
    theFPD[4].ParseNewText(varTexto, 'uv', '', True, True);
    for i:= 0 to numOfCPu-1 do
      theFPDClones[i,4].ParseNewText(varTexto, 'uv', '', True, True);
    inputStrs[4]:= varTexto;

    varTexto:= Trim(uMinEdit.Text);
    uMinEdit.Text:= varTexto;
    if (varTexto = '') then
      raise Exception.Create(eBlankField);
    ReadConstantField(varTexto, min2D[1]);
    inputStrs[5]:= varTexto;

    varTexto:= Trim(uMaxEdit.Text);
    uMaxEdit.Text:= varTexto;
    if (varTexto = '') then
      raise Exception.Create(eBlankField);
    ReadConstantField(varTexto, max2D[1]);
    inputStrs[6]:= varTexto;

    if max2D[1] <= min2D[1] then
      raise Exception.Create(euMinLessThanuMax);

    varTexto:= Trim(vMinEdit.Text);
    vMinEdit.Text:= varTexto;
    if (varTexto = '') then
      raise Exception.Create(eBlankField);
    ReadConstantField(varTexto, min2D[2]);
    inputStrs[7]:= varTexto;

    varTexto:= Trim(vMaxEdit.Text);
    vMaxEdit.Text:= varTexto;
    if (varTexto = '') then
      raise Exception.Create(eBlankField);
    ReadConstantField(varTexto, max2D[2]);
    inputStrs[8]:= varTexto;

    if max2D[2] <= min2D[2] then
      raise Exception.Create(evMinLessThanvMax);

    for i:= 1 to 2 do begin
      edges2D[i]:= Max2D[i] - Min2D[i];
      halfEdges2D[i]:= edges2D[i] / 2;
      center2D[i]:= (Max2D[i] + Min2D[i]) / 2;
    end;
    span2D:= hypot(edges2D[1], edges2D[2]);
    radius2D:= span2D / (kWidthMainBox + kHeightMainBox);
    maxSegLen:= 5 * radius2D;
    cX:= kWidthMainBox / edges2D[1];
    cY:= kHeightMainBox / edges2D[2];
    qx:= kWidthMainBox / 2 - cX * center2D[1];
    qy:= kHeightMainBox / 2 + cY * center2D[2];
    dirty:= True;
    alreadyParsed:= True;
    bm.canvas.FillRect(rctMain);
    mainBox.canvas.FillRect(rctMain);
    InitializeLines;
    InitializeLoci;
    crossDrawn:= False;
    busy:= False;
    drawOrthonLine:= False;
    stop:= False;
    sc:= 0;
    ellFactor:= 1;
    lockPtData:= False;
    backgroundType:= -1;
    PrintBackground:= False;
    InfoBar.SimpleText:= '';
    ManageMenus;
    result:= True;
  except on E: Exception do begin
      result:= False;
      Beep(100, 100);
      InfoBar.SimpleText:= m;
      AlreadyParsed:= False;
    end;
  end;
end;

procedure TParam4.RedrawLines;
begin
  if not alreadyParsed then
    Exit;
  DrawLines(mainBox.canvas, rctMain);
end;

procedure TParam4.RedrawLoci;
var
  i: Integer;
begin
  if not alreadyParsed then
    Exit;
  for i:= kParabolicLocus to kZeroDiscrStrongPpalLocus do
    DrawLocus(mainBox.canvas, rctMain, false, i);
end;

procedure TParam4.ShowData(param: VectorsAndUtil.Vector2; halfMajorAxis, halfMinorAxis,
  lengthH, depth,
  alpha, KGAuss, discrNormDev, discrLatDev, discrStrPpal: Double);
begin
  with FunctionValues1.Strings do begin
    Strings[0]:= 'u =' + FloatToStrF(param[1], ffFixed, 14, 10);
    Strings[1]:= 'v =' + FloatToStrF(param[2], ffFixed, 14, 10);
    Strings[2]:= 'Maj Axis =' + FloatToStrF(halfMajorAxis, ffFixed, 14, 10);
    Strings[3]:= 'Min Axis =' + FloatToStrF(halfMinorAxis, ffFixed, 14, 10);
    Strings[4]:= 'Norm of H  =' + FloatToStrF(lengthH, ffFixed, 14, 10);
    Strings[5]:= 'H depth =' + FloatToStrF(depth, ffFixed, 14, 10);
    if alpha = kTheError then
      Strings[6]:= 'Sin <(H,E) ='
    else
      Strings[6]:= 'Sin <(H,E) =' + FloatToStrF(alpha, ffFixed, 14, 10);
    Strings[7]:= 'Gauss Curv =' + FloatToStrF(KGAuss, ffFixed, 14, 10);
    Strings[8]:= 'Discr Extr Norm Dev=' + FloatToStrF(discrNormDev, ffFixed, 14,
      10);
    Strings[9]:= 'Discr Extr Lat Dev=' + FloatToStrF(discrLatDev, ffFixed, 14,
      10);
    Strings[10]:= 'Discr Str Ppal Dirs=' + FloatToStrF(discrStrPpal, ffFixed, 14,
      10);
  end;
  FunctionValues1.Update;
end;

procedure TParam4.ToDisk;
var
  sl: TStringList;
  i: Integer;
begin
  sl:= TStringList.Create;
  try
    sl.BeginUpdate;
    sl.Clear;
{$O-}
    for i:= -3 to kLastStr do
      if inputStrs[i]='' then
        sl.Append(' ') {don't append an empty string because then nothing is added}
      else
        sl.Append(inputStrs[i]);
    sl.EndUpdate;
{$O+}
    sl.SaveToFile(fileName);
  finally
    FreeAndNil(sl);
  end;
end;

end.

