unit InfByteArrays;
{Parametricas4, workshop for the geometry of surfaces in R^4.


MIT License

Copyright (c) [2010 - 2018] [Angel Montesinos-Amilibia], email montesin@uv.es

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
}

interface

uses
  Classes, SysUtils;
resourcestring
  eWrongAttemptToWrite = 'Wrong attempt to write on an open array';
  eWrongAttemptToDelete= 'Wrong attempt to delete from an open array';


type
  ByteSet= Set of Byte;

  ByteArray= packed array of Byte;

  InfByteArray= record
     h, bigH: Integer;
    arr: ByteArray;
  end;


  InfBArray=  record
    h, bigH: Integer;
    arr: PByte;
  end;


  {the next procedures beginning with IB, and with assume that the array to be chaned has
   enough room}
  procedure IBInsert(src:  array of Byte; var dst: InfBArray; frst: Integer);
  procedure IBDelete(var dst: InfBArray; first, last: Integer);
  procedure IBAppend(var b: InfBArray; a: array of Byte); overload;
  procedure IBAppend(var b: InfBArray; a: Byte); overload;
  procedure IBPrepend(var b: InfBArray; a: array of Byte);

  procedure IBAInsert(src:  array of Byte; var dst: InfByteArray; frst: Integer); overload;
  procedure IBAInsert(src:  InfByteArray; var dst: InfByteArray; frst: Integer); overload;
  procedure IBADelete(var dst: InfByteArray; first, last: Integer);
  procedure IBAAppend(var b: InfByteArray; a: array of Byte); overload;
  procedure IBAAppend(var b: InfByteArray; a: Byte); overload;
  procedure IBAPrepend(var b: InfByteArray; a: array of Byte);


  procedure AppendBA(var b: ByteArray; a: array of Byte); overload;
  procedure AppendBA(var b: ByteArray; a: Byte); overload; 
  procedure PrependBA(var b: ByteArray; a: array of Byte); overload;
  procedure PrependBA(var b: ByteArray; a: Byte); overload;
  procedure InsertBA(src: array of byte; var dst: ByteArray; lstDest, frst: Integer);
  procedure DeleteBA(var bA: PByte; var hbA: Integer; first, last: Integer);  overload;
  procedure DeleteBA(var bA: ByteArray; var hbA: Integer; first, last: Integer);  overload;
  function ExtractBA(src: PByte; first, amount: Integer): ByteArray;

 
  
  

implementation
  
{$O-} 
  
 procedure IBAppend(var b: InfBArray; a: array of Byte);
 var 
    i: Integer;    
  begin
    if  High(a) + b.h + 1 > b.bigH then 
      raise exception.create(eWrongAttemptToWrite); 
    for i:= 0 to High(a) do
      b.arr[b.h+1+i]:= a[i];
    b.h:= b.h + High(a) + 1;
  end;
 
 procedure IBAppend(var b: InfBArray; a: Byte);
  begin
    if b.h >= b.bigH then 
      raise exception.create(eWrongAttemptToWrite); 
    b.arr[b.h+1]:= a;
    inc(b.h);
  end;

 procedure IBDelete(var dst: InfBArray; first, last: Integer);
 {Deletes from 'dst' the entries 'dst.arr[first]..dst.arr[last]' and updates
  the value of 'dst.h'}
 var
   i, step: Integer;
 begin
   if (first < 0) or (last  > dst.h) or (last < first) then
     raise exception.Create(eWrongAttemptToDelete);
   step:= last - first + 1;
   for i:= last+1 to dst.h do
     dst.arr[i-step]:= dst.arr[i];
   dst.h:= dst.h - step;
 end; 
 
 procedure IBPrepend(var b: InfBArray; a: array of Byte);
 var
   i, step: Integer;
 begin
   step:= high(a)+1;
   if b.h + step > b.bigH then
     raise exception.Create(eWrongAttemptToDelete);
   if step > 0 then begin
     for i:= b.h downto 0 do
       b.arr[i+step]:= b.arr[i];
     for i:= 0 to High(a) do
       b.arr[i]:= a[i];
     inc(b.h, step);
   end;
 end;
 
 procedure IBInsert(src: array of Byte; var dst: InfBArray; frst: Integer);
 {Inserts 'src' into 'dst.arr' so that all the old values 'dst.arr[frst]..
  dst.arr[dst.h]' are copied to 'dst.arr[frst+high(src)+1]...'. The
  value of 'dst.h' is updated.}
 var
   i, step: Integer;                                                 
 begin
   step:= high(src) + 1;
   if step <= 0 then
     Exit;
   if (frst < 0) or (frst > dst.h+1) or (step + dst.h > dst.bigH) then
     raise exception.create(eWrongAttemptToWrite);
   for i:= dst.h downto frst do
     dst.arr[i+step]:= dst.arr[i];
   for i:= 0 to high(src) do
     dst.arr[frst+i]:= src[i];
   dst.h:= dst.h + step;
 end;



 procedure IBAInsert(src:  array of Byte; var dst: InfByteArray; frst: Integer);
 {Inserts 'src' into 'dst.arr' so that all the old values 'dst.arr[frst]..
  dst.arr[dst.h]' are copied to 'dst.arr[frst+high(src)+1]...'. The
  value of 'dst.h' is updated.}
 var
   i, step: Integer;
 begin
   step:= high(src) + 1;
   if step <= 0 then
     Exit;
   if (frst < 0) or (frst > dst.h+1) or (step + dst.h > dst.bigH) then
     raise exception.create(eWrongAttemptToWrite);
   for i:= dst.h downto frst do
     dst.arr[i+step]:= dst.arr[i];
   for i:= 0 to high(src) do
     dst.arr[frst+i]:= src[i];
   dst.h:= dst.h + step;
 end;

  procedure IBAInsert(src:  InfByteArray; var dst: InfByteArray; frst: Integer);
  {Inserts 'src' into 'dst.arr' so that all the old values 'dst.arr[frst]..
  dst.arr[dst.h]' are copied to 'dst.arr[frst+src.h+1]...'. The
  value of 'dst.h' is updated.}
  var
   i, step: Integer;
 begin
   step:= src.h + 1;
   if step <= 0 then
     Exit;
   if (frst < 0) or (frst > dst.h+1) or (step + dst.h > dst.bigH) then
     raise exception.create(eWrongAttemptToWrite);
   for i:= dst.h downto frst do
     dst.arr[i+step]:= dst.arr[i];
   for i:= 0 to src.h do
     dst.arr[frst+i]:= src.arr[i];
   dst.h:= dst.h + step;
 end;

  procedure IBADelete(var dst: InfByteArray; first, last: Integer);
  {Deletes from 'dst' the entries 'dst.arr[first]..dst.arr[last]' and updates
  the value of 'dst.h'}
 var
   i, step: Integer;
 begin
   if (first < 0) or (last  > dst.h) or (last < first) then
     raise exception.Create(eWrongAttemptToDelete);
   step:= last - first + 1;
   for i:= last+1 to dst.h do
     dst.arr[i-step]:= dst.arr[i];
   dst.h:= dst.h - step;
 end;


 procedure IBAAppend(var b: InfByteArray; a: array of Byte);
  var
    i: Integer;
  begin
    if  High(a) + b.h + 1 > b.bigH then
      raise exception.create(eWrongAttemptToWrite);
    for i:= 0 to High(a) do
      b.arr[b.h+1+i]:= a[i];
    b.h:= b.h + High(a) + 1;
  end;

 procedure IBAAppend(var b: InfByteArray; a: Byte);
  begin
    if b.h >= b.bigH then
      raise exception.create(eWrongAttemptToWrite);
    b.arr[b.h+1]:= a;
    inc(b.h);
  end;

 procedure IBAPrepend(var b: InfByteArray; a: array of Byte);
  var
   i, step: Integer;
 begin
   step:= high(a)+1;
   if b.h + step > b.bigH then
     raise exception.Create(eWrongAttemptToDelete);
   if step > 0 then begin
     for i:= b.h downto 0 do
       b.arr[i+step]:= b.arr[i];
     for i:= 0 to High(a) do
       b.arr[i]:= a[i];
     inc(b.h, step);
   end;
 end;




 
procedure AppendBA(var b: ByteArray; a: array of Byte);
var 
  i, h: Integer;    
begin
  h:= High(b);
  if  High(a) >= 0 then begin
    SetLength(b, h + High(a) + 2);
    for i:= 0 to High(a) do
      b[h+1+i]:= a[i];
  end;
end;  

procedure AppendBA(var b: ByteArray; a: Byte);    
begin
  SetLength(b, High(b) + 2);
  b[High(b)]:= a;
end;    
 
 
 procedure InsertBA(src: array of byte; var dst: ByteArray; lstDest, frst: Integer);
 {'hbA' denotes the last meaningful index of 'dst'. Inserts 'src' into 'dst'
  so that all the old values 'dst[frst]..dst[high(dst)]' are copied to 
  'dst[frst+high(src)+1]...'.}
 var
   i, step: Integer;
 begin
   if High(src) < 0 then
     Exit;
   if (frst < 0) or (frst > lstDest+1) then
     raise exception.create(eWrongAttemptToWrite);
   step:= high(src) + 1;
   SetLength(dst, lstDest + 1 +step);
   if frst <= lstDest then
     for i:= lstDest downto frst do
       dst[i+step]:= dst[i];
   for i:= 0 to high(src) do
     dst[frst+i]:= src[i];
 end;
 
 procedure DeleteBA(var bA: PByte; var hbA: Integer; first, last: Integer); overload;
 {Deletes from 'bA' the entries 'bA[first]..bA[last]' and modifies the value
  of 'hbA', having in mind that 'bA[hbA]' must be the last meaningful entry
  of 'bA'}
 var
   i, step: Integer;
 begin
   if (last < first) or (first< 0) or (last  > hbA)  then
     raise exception.Create(eWrongAttemptToDelete);
   step:= last - first + 1;
   for i:= last+1 to hbA do
     bA[i-step]:= bA[i];
   hbA:= hbA - step;
 end;

 procedure DeleteBA(var bA: ByteArray; var hbA: Integer; first, last: Integer);  overload;
 var
   i, step: Integer;
 begin
   if (last < first) or (first< 0) or (last  > hbA)  then
     raise exception.Create(eWrongAttemptToDelete);
   step:= last - first + 1;
   for i:= last+1 to hbA do
     bA[i-step]:= bA[i];
   hbA:= hbA - step;
 end;
 
 function ExtractBA(src: PByte; first, amount: Integer): ByteArray;
 {produces the ByteArray whose content is 'src[first]..src[first+amount-1]'} 
 var
   i: Integer;
 begin
   if amount < 1 then
     result:= nil
   else begin
     SetLength(result, amount);
     for i:= first to first+amount-1 do
       result[i-first]:= src[i];
   end;
 end;
 
 procedure PrependBA(var b: ByteArray; a: array of Byte);
 var
   i, step, h: Integer;
 begin
   step:= high(a)+1;
   h:= High(b);
   if step > 0 then begin
     SetLength(b, h + 1 + step);
   for i:= h downto 0 do
     b[i+step]:= b[i];
   for i:= 0 to High(a) do
     b[i]:= a[i];
   end;
 end; 
 
 procedure PrependBA(var b: ByteArray; a: Byte);
 var
   i, h: Integer;
 begin
   h:= High(b);
   SetLength(b, h + 2);
   for i:= h downto 0 do
     b[i+1]:= b[i];
   b[0]:= a;
 end; 
 
{$O+} 

end.

