
{-----------------------------------------------------------------------------
 Unit Name: VectorsAndUtil
 Project Name : Parametricas4
 Author:    Angel Montesinos Amilibia
 Dates: 2003-2009
 Purpose: global constans and types.  Routines in linear algebra
-----------------------------------------------------------------------------}
{

Parametricas4, workshop for the geometry of surfaces in R^4.

MIT License

Copyright (c) [2003 - 2018] [Angel Montesinos-Amilibia], email montesin@uv.es

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.


}



unit VectorsAndUtil;

interface
  uses
    Windows, Graphics;

const
  kIncr2Points = 1.4e-7;
  k2incr2Points= 2.8e-7;

  kIncr4Points= 0.0002;
  k2Incr4Points= 2*kIncr4Points;
  k12Incr4Points = 12*kIncr4Points;

  kTheError = 1511724757.0;

  pi :Double =3.1415926535897932384;

  kPPLine2D = 20;
  kMaxPPLine2D = 50;
  kMaxNumFlags = 4 * kMaxPPLine2D;

type
  Vector2 = array[1..2] of Double;
  Vector3 = array[1..3] of Double;
  Vector4 = array[1..4] of Double;

  Vector2Single = array[1..2] of Single;
  Vector3Single = array[1..3] of Single;

  Vector2Int = array[1..2] of Integer;
  Vector3Int = array[1..3] of Integer;

  Segment2 = array[1..2] of Vector2;
  Segment3 = array[1..2] of Vector3;

  matrix33 = array[1..3] of Vector3;
  matrix44 = array[1..4] of Vector4;

  ScreenSegment = array[1..2] of TPoint;


  MapR1R2= procedure (var par: Vector2; t: Double) of object;
  MapR1R1= function (t: Double): Double of object;
  MapR2R1= function (const p : Vector2) : Double of object;

  InRange2DFunction = function(var depth: Double; p: Vector2): Boolean of object;
  StoreSegmentProcedure2D = procedure(currentX: Vector2; key:
    Integer) of object;

  function Area(u,v: Vector2): Double;
  function CrossProd(u: Matrix44): Vector4;
  function Determinant44(mc : matrix44): Double;
  function EuclCrossProd(u,v: Vector3): Vector3;
  function EuclDistante(x,y: Vector4): Double; overload;
  function EuclDistance(x,y: Vector3): Double; overload;
  function EuclDistance(x,y: Vector2): Double; overload;
  function EuclEnergy(x: Vector2): Double; overload;
  function EuclEnergy(x: Vector3): Double; overload;
  function EuclEnergy(x: Vector4): Double; overload;
  function EuclInnerProd(u,v: Vector2): Double; overload;
  function EuclInnerProd(u,v: Vector3): Double; overload;
  function EuclInnerProd(u,v: Vector4): Double; overload;
  function EuclNorm(v: Vector2): Double; overload;
  function EuclNorm(v: Vector3): Double; overload;
  function EuclNorm(v: Vector4): Double; overload;
  function FormatFloatString(const x: Double; const totalStrLength: Integer): string;
  function MyArctan2(y,x: Double): Double;
  function SqrEuclDistance(x,y: Vector3): Double;


implementation

uses
  SysUtils, Math;

function Area(u,v: Vector2): Double;
begin
 result:= u[1]*v[2] - u[2]*v[1];
end;

function CrossProd(u: Matrix44): Vector4;
{Taken from Steven Richard Hollasch, Thesis,
http://research.microsoft.com/~hollasch/thesis/default.htm}
var
  a, b, c, d, e, f: Double;
begin
  a:= (u[2, 1] * u[3, 2]) - (u[2, 2] * u[3, 1]);
  b:= (u[2, 1] * u[3, 3]) - (u[2, 3] * u[3, 1]);
  c:= (u[2, 1] * u[3, 4]) - (u[2, 4] * u[3, 1]);
  d:= (u[2, 2] * u[3, 3]) - (u[2, 3] * u[3, 2]);
  e:= (u[2, 2] * u[3, 4]) - (u[2, 4] * u[3, 2]);
  f:= (u[2, 3] * u[3, 4]) - (u[2, 4] * u[3, 3]);

  result[1]:= (u[1, 2] * f) - (u[1, 3] * e) + (u[1, 4] * d);
  result[2]:= -(u[1, 1] * f) + (u[1, 3] * c) - (u[1, 4] * b);
  result[3]:= (u[1, 1] * e) - (u[1, 2] * c) + (u[1, 4] * a);
  result[4]:= -(u[1, 1] * d) + (u[1, 2] * b) - (u[1, 3] * a);
end;

function Determinant44(mc : matrix44): Double;
{Only the entries  mc[i,j]  with i,j <5 are used}
var
  i, j, k, iBest : Integer;
  m : matrix44;
  aVector : vector4;
  aux : Double;
begin
  result := 1;
  m := mc;
  for k := 1 to 2 do begin
    iBest := 0;
    aux := 0;
    for i := k to 4 do
    if abs(m[i, k]) > aux then begin
      aux := abs(m[i, k]);
      iBest := i;
    end;
    if iBest = 0 then begin
      result := 0;
      Exit;
    end;
    if iBest > k then begin
      aVector := m[iBest];
      m[iBest] := m[k];
      m[k] := aVector;
      result := -result * m[k, k];
    end
    else
      result := result * m[k, k];
      for j := k + 1 to 4 do
        m[k, j] := m[k, j] / m[k, k];
      for i := k + 1 to 4 do
        for j := k + 1 to 4 do
          m[i, j] := m[i, j] - m[k, j] * m[i, k];
  end;
  result := result * (m[3, 3] * m[4, 4] - m[3, 4] *
        m[4, 3]);
end;

function EuclCrossProd(u,v: Vector3): Vector3;
begin
  result[1]:= u[2]*v[3] - u[3]*v[2];
  result[2]:= u[3]*v[1] - u[1]*v[3];
  result[3]:= u[1]*v[2] - u[2]*v[1];
end;

function EuclDistante(x,y: Vector4): Double;
begin
  result:= sqrt(sqr(x[1]-y[1]) + sqr(x[2]-y[2]) + sqr(x[3]-y[3])+sqr(x[4]-y[4]));
end;

function EuclDistance(x,y: Vector3): Double;
begin
  result:= sqrt(sqr(x[1]-y[1]) + sqr(x[2]-y[2]) + sqr(x[3]-y[3]));
end;

function EuclDistance(x,y: Vector2): Double;
begin
  result:= sqrt(sqr(x[1]-y[1]) + sqr(x[2]-y[2]));
end;

function EuclEnergy(x: Vector2): Double; overload;
begin
  result:= sqr(x[1])+ sqr(x[2]);
end;

function EuclEnergy(x: Vector3): Double; overload;
begin
  result:= sqr(x[1])+ sqr(x[2])+sqr(x[3]);
end;

function EuclEnergy(x: Vector4): Double;
begin
  result:= sqr(x[1])+ sqr(x[2])+sqr(x[3])+sqr(x[4]);
end;

function EuclInnerProd(u,v: Vector2): Double; overload;
begin
  result:= u[1]*v[1] + u[2]*v[2];
end;

function EuclInnerProd(u,v: Vector3): Double; overload;
begin
  result:= u[1]*v[1] + u[2]*v[2] + u[3]*v[3];
end;


function EuclInnerProd(u,v: Vector4): Double; overload;
begin
  result:= u[1]*v[1] + u[2]*v[2] + u[3]*v[3] + u[4]*v[4];
end;

function EuclNorm(v: Vector2): Double; overload;
begin
  result:= sqrt(sqr(v[1]) + sqr(v[2]));
end;

function EuclNorm(v: Vector3): Double; overload;
begin
  result:= sqrt(sqr(v[1]) + sqr(v[2]) + sqr(v[3]));
end;

function EuclNorm(v: Vector4): Double; overload;
begin
  result:= sqrt(sqr(v[1]) + sqr(v[2]) + sqr(v[3]) + sqr(v[4]));
end;

function FormatFloatString(const x: Double; const totalStrLength: Integer): string;
{Gives the string of 'x' with 'totalStrLength' characters unless this
 number be greater than 15}
  var
    sigDig: Integer;
begin
  result:= FloatToStrF(x, ffFixed, totalStrLength+2,15);
  sigDig:= Length(result);
  if sigDig > totalStrLength  then
    sigDig:= totalStrLength;
  result:= Copy(result,1,sigDig);
end;

function MyArctan2(y,x: Double): Double;
begin
  if abs(x) > 1.0e-15 then
    result:= arctan2(y,x)
  else begin
    if (y > 0) and (y < 1.0e10) then
      result:= pi/2
    else if ( y < 0 ) and (y > -1.0e10) then
      result:= -pi/2
    else
      result:= 0;
  end;
end;

function SqrEuclDistance(x,y: Vector3): Double;
begin
  result:= sqr(x[1]-y[1]) + sqr(x[2]-y[2]) + sqr(x[3]-y[3]);
end;




END.
 